(* Timbuk - A Tree Automata Library (for Ocaml)
   Copyright (C) 2000-2001 Thomas Genet and Valrie Viet Triem Tong

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Last modification 28/02/01
*)

type ('sym, 'var) term_const=
    Var of 'var
  | Const of 'sym
  | Fsym of ('sym * (('sym, 'var) term_const) list)
  | Special of ('sym, 'var) term_const

    
module type PRINTABLE_TYPE =
  sig
    type t
    val to_string: t -> string
    val of_string: string -> t
    val equal: t -> t -> bool
  end

module type STATE_CONTENT_TYPE =
  sig
    type t
    val to_string: t -> string
    val of_string: string -> t
    val equal: t -> t -> bool
    val empty: t
    val inter: t -> t -> t
  end

module type ALPHABET_TYPE =
  sig
    type t
    type symbol

    exception Symbol_not_in_alphabet of string
    val new_alphabet: t
    val union_fast : t -> t -> t
    val union: t -> t -> t
    val disjoint: t -> t -> bool
    val add_symbol: symbol -> int -> t -> t
    val get_arity: symbol -> t -> int
    val occur: symbol -> t -> bool
    val to_string: t -> string
    val to_list: t -> (symbol * int) list
    val parse: Genlex.token Stream.t -> t
  end


module type VARIABLE_SET_TYPE =
  sig
    type t
    type variable
    val empty: t
    val is_empty: t -> bool
    val mem: variable -> t -> bool
    val to_string: t -> string
    val parse: Genlex.token Stream.t -> t
  end


module type TERM_TYPE= 
sig

  type symbol
  type variable
  type alphabet
  type variable_set
  type term= (symbol, variable) term_const

  type t= term

  type substitution= (variable * term) list
			
  exception Terms_do_not_match of string * string
  exception Badly_formed_term of string
  exception Undefined_symbol of string 
  exception Bad_operation_on_special_term of string
  exception Bad_operation_on_var of string
  
			
  val equal : t -> t -> bool
  val depth : t -> int
  val to_string : t -> string
  val top_symbol : t -> symbol
  val list_arguments : t -> t list
  val is_variable : t -> bool
  val is_constant : t -> bool
  val is_special : t -> bool
  val get_special : t -> t
  val map : f1:(symbol -> symbol) -> f2:(t -> t) -> t1:t -> t
  val list_leaves : t -> t list
  val list_variables : t -> variable list
  val is_ground : t -> bool
  val is_linear : t -> bool
  val list_special : t1:t -> t list
  val check_special : alphabet -> alphabet -> t1:t -> t
  val check : alphabet -> t -> t
  val replace_special : l:((t * t) list) -> t3:t -> t
  val apply : substitution  -> t -> t
  val apply_several : l:(substitution list) -> t1:t -> t list
  val parse_special :
    alphabet ->
        alphabet ->
        variable_set -> Genlex.token Stream.t -> t
  val parse :
      alphabet ->
        variable_set -> Genlex.token Stream.t -> t
	
  val parse_ground :
    alphabet ->
        Genlex.token Stream.t -> t
	
  val parse_ground_special :
    alphabet ->
        alphabet ->
	Genlex.token Stream.t -> t
	
  val coherent : substitution -> substitution
  val matching : t -> t -> substitution
end 


module type STATE_SET_TYPE =
  sig
    type alphabet
    type symbol
    type state_content
    type t
    type state
    val default_prod_symbol: symbol
    val state_product: state -> state -> state
    val empty: t
    val is_empty: t -> bool
    val first: t -> state
    val remainder: t -> t
    val add: state -> t -> t
    val add_verb: state -> state_content -> t -> t
    val add_all: state list -> t -> t
    val add_all_verb : state list -> t -> t -> t
    val mem: state -> t -> bool
    val inter: t -> t -> t
    val union: t -> t -> t
    val minus: t -> t -> t
    val all_mem: state list -> t -> bool
    val list_to_set: state list -> t
    val to_string: t -> string
    val state_description : state -> t -> state_content
    val symbolic_product : t -> t -> t
    val to_string_verb: t -> string
    val parse_ops: Genlex.token Stream.t -> alphabet
    val parse: alphabet -> alphabet -> Genlex.token Stream.t -> t
    val parse_verb: alphabet -> alphabet -> Genlex.token Stream.t -> t
  end



module type TRS_TYPE=
sig
  type variable_set
  type alphabet
  type term
  type ruleSystem
  type t= ruleSystem
  type rule

  val empty : t
  val new_rule : term -> term -> rule
  val is_empty : t -> bool
  val first : t -> rule
  val remainder : t -> t
  val rhs : rule -> term
  val lhs : rule -> term
  val rule_equal : rule -> rule -> bool
  val is_left_linear : rule -> bool
  val is_right_linear : rule -> bool
  val is_linear : rule -> bool
  val non_linear_lhs : t -> term list
  val add : rule -> t -> t
  val mem : rule -> t -> bool
  val union : t -> t -> t
  val inter : t -> t -> t
  val add_fast : rule -> t -> t
  val union_fast : t -> t -> t
  val to_list : t -> rule list
  val list_to_trs : rule list -> t
  val rule_to_string : rule -> string
  val to_string : t -> string

  val check_special_rule :
    rule -> alphabet -> alphabet -> rule
  val check_special :
    t -> alphabet -> alphabet -> t
  val check_rule : rule -> alphabet -> rule
  val check : t -> alphabet -> t
  val parse_special_rule :
    alphabet ->
      alphabet ->
        variable_set -> Genlex.token Stream.t -> rule

  val parse_rule :
    alphabet ->
      variable_set -> Genlex.token Stream.t -> rule
  
  val parse :
      alphabet ->
        variable_set -> Genlex.token Stream.t -> t
  val parse_special :
    alphabet ->
      alphabet ->
        variable_set -> Genlex.token Stream.t -> t

  val parse_ground_special :
    alphabet ->
      alphabet -> Genlex.token Stream.t -> t
  val parse_ground_special_rule :
    alphabet -> alphabet -> Genlex.token Stream.t -> rule

  val rewrite_top_once : t -> term -> term
  val left_inner_norm : t -> term -> term
  val left_inner_norm_special :
    t -> term -> term

  val left_inner_norm_special_system :
    t -> t -> t
  val bottom_up_norm : t -> term -> term
end 

  
module type AUTOMATA_TYPE=
sig 
  exception Not_a_state of string
  exception Not_in_folder
  exception Multiply_defined_symbol of string
  exception Linearity_problem of string
  exception Normalisation_problem of string
    
  type symbol
  type alphabet
  type term
  type rule
  type substitution
  type state
  type state_set

  type transition_table
  type tree_automata

  type t = tree_automata
  type ('a, 'b) folder


  val get_alphabet : t -> alphabet
  val get_state_ops : t -> alphabet
  val get_states : t -> state_set
  val get_final_states : t -> state_set
  val get_transitions : t -> transition_table
  val get_prior : t -> transition_table
  val make_automaton:
    main_alphabet: alphabet ->
      state_operators: alphabet ->
        state_set ->
          final_states: state_set ->
            transitions: transition_table -> 
	      prioritary_transitions: transition_table -> t

  val modify_final : t -> state_set -> t
  val is_empty : t -> bool
  val make_state : symbol -> state
  val make_state_config : state -> term
  val new_trans : symbol -> state list -> state -> rule
  val is_state_config : term -> bool
  val lhs : rule -> term
  val rhs : rule -> term
  val top_symbol :
    rule -> symbol
  val state_label : term -> state
  val is_normalized : rule -> bool
  val list_state : rule -> state list
  val to_string : t -> string
  val states_of_transitions :
    transition_table -> state_set
  val parse : alphabet -> Genlex.token Stream.t -> t

  val accessibility_cleaning : t -> t
  val utility_cleaning : t -> t
  val clean : t -> t
  val rewrite_state_labels :
    t -> transition_table -> t

  val automatic_renum : t -> t
  val union : t -> t -> t
  val simplify : t -> t

  val inter : t -> t -> t
  val is_language_empty : t -> bool 

  val normalize_epsilon :
    term ->
      term ->
	transition_table -> transition_table

  val normalize :
    transition_table ->
      transition_table ->
        string -> int -> transition_table * int * alphabet

  val matching :
    term ->
      term ->
        (symbol, (state, rule list) folder) folder ->
          substitution list

  val disjointness_constraint :
    term list ->
      (symbol, (state, rule list) folder) folder ->
        term list list 

  val is_recognized_into :
    term ->
      term ->
        (symbol, (state, rule list) folder) folder -> bool

  val run : 
    term ->
      state ->
	t -> bool
    
  val is_covered :
    term ->
      term ->
        (symbol, (state, rule list) folder) folder -> bool 

  val determinise : t -> t

  val make_complete : t -> t

  val make_red_automaton :
    alphabet -> transition_table -> t

  val inverse : t -> t
  val subtract : t -> t -> t
  val is_included : t -> t -> bool 
  val nf_automaton :
    alphabet -> transition_table -> t

  (* functions for folders of transitions *)

  val folder_assoc : criteria:'a -> folder:('a, 'b) folder -> 'b
  val folder_add :
    trans:'a -> criteria:'b -> folder:(('b , 'a list) folder) -> ('b , 'a list) folder
  val folder_replace : 'a -> 'b -> ('b , 'a) folder -> ('b , 'a) folder
  val bi_folder_add :
    trans:'a ->
      criteria1:'b ->
        criteria2: 'c ->
          bi_folder: ('b , ('c , 'a list) folder) folder -> ('b , ('c , 'a list) folder) folder
  val bi_folder_mem :
    rule ->
      (symbol ,
       (state , rule list) folder) folder -> bool
  val bi_folder_add_trans_list :
    rule list ->
      (symbol ,
       (state , rule list) folder) folder ->
        (symbol ,
       (state , rule list) folder) folder
  val configs_from_symbol_to_state :
    symb: 'a ->
      q: 'b ->
        ('a , ('b , rule list) folder) folder ->
          term list
  val folder_flatten : ('a , 'b list) folder -> 'b list

  val bi_folder_flatten : ('a , ('b , 'c list) folder) folder -> 'c list
  val transitions_by_symbol :
    rule list ->
      (symbol , rule list) folder
  val transitions_by_state :
    rule list ->
      (state , rule list) folder
  val transitions_by_state_by_symbol :
    rule list ->
      (symbol ,
       (state , rule list) folder) folder
end 


module type AUTOMATON_SAVING_TYPE=
sig 
  type automaton
  val save_automaton : a:automaton -> aut_name: string ->  file_name:string -> unit
end


module type GAMMA_TYPE=
sig
  type alphabet
  type automaton
  type transition_table
  type transition
  type gamma_content
  type variable_set
  type ('a, 'b) folder
  type state
  type symbol
 
  val out_chan : out_channel ref
  val in_chan : in_channel ref

  val empty : gamma_content
  val add_prior_to_gamma :
    transition_table -> gamma_content -> gamma_content
  val set_prior_in_gamma :
    transition_table -> gamma_content -> gamma_content
  val get_prior : gamma_content -> transition_table

  val go: unit -> bool

  val gamma :
    transition_table ->
      (symbol , (state, transition list) folder) folder ->
        gamma_content ->
          alphabet ->
            alphabet ->
              int -> transition_table * gamma_content * alphabet * int

  val parse :
    alphabet ->
      variable_set -> Genlex.token Stream.t -> gamma_content

  val to_string :
    gamma_content -> string
end

  
(* <= for lists *)

let subeq_list l1 l2 = 
  List.for_all (function a -> List.mem a l2) l1

