(* Timbuk - A Tree Automata Library (for Ocaml)
   Copyright (C) 2000-2001 Thomas Genet and Valrie Viet Triem Tong

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Last modification 28/02/01
*)

open Common

module Gamma 
  (Symbol_type: PRINTABLE_TYPE)(Alphabet_type: ALPHABET_TYPE with type symbol= Symbol_type.t)
  (Variable_set_type: VARIABLE_SET_TYPE)
  (Term: TERM_TYPE with type symbol= Symbol_type.t 
		   and type alphabet= Alphabet_type.t
		   and type variable_set= Variable_set_type.t)
  (TRS_type: TRS_TYPE with type alphabet= Alphabet_type.t 
		      and type term= Term.t
		      and type variable_set= Variable_set_type.t)
  (State_set_type: STATE_SET_TYPE with type state= Term.t
				  and type alphabet= Alphabet_type.t and type symbol= Symbol_type.t)
  (Automata_type: AUTOMATA_TYPE with type symbol= Symbol_type.t 
				and type alphabet= Alphabet_type.t 
				and type term= Term.term 
				and type transition_table= TRS_type.t 
				and type rule= TRS_type.rule)=

struct
    module State_type= Term
    module Trans_type= TRS_type

(*  channels for user interaction *)

    let out_chan= ref stdout
    let in_chan= ref stdin

    type automaton = Automata_type.t
    type alphabet = Alphabet_type.t	
    type state = Automata_type.state
    type transition_table = Trans_type.t
    type transition = Trans_type.rule
    type variable_set= Variable_set_type.t
    type subst= Term.substitution
    type ('a, 'b) folder= ('a, 'b) Automata_type.folder
    type symbol= Symbol_type.t
    type rule= Trans_type.rule
		  
    type norm_rule = {pat:transition; modifs: transition_table}
    type gamma_content= {prior: transition_table; new_states:alphabet; 
			 norm_rules: norm_rule list; strategy: string list}
		
    exception Multiply_defined_symbol of string
    exception Error_in_parsing of string
    exception Badly_formed_term of string
    exception Strategy_error of string 

    (* The empty gamma approximation content with the default strategy ! *)
 
    let empty= {prior=Trans_type.empty; new_states=Alphabet_type.new_alphabet; norm_rules=[];
	       strategy= ["prior"; "norm_rules"; "manual_conf"; "auto_conf"]}

    let add_prior_to_gamma ltrans (g: gamma_content)= {prior= Trans_type.union ltrans g.prior;
						       new_states= g.new_states;
						       norm_rules= g.norm_rules;
						       strategy=g.strategy}
    let set_prior_in_gamma ltrans (g: gamma_content)= {prior= ltrans; 
						       new_states= g.new_states;
						       norm_rules= g.norm_rules;
						       strategy=g.strategy}
    let get_prior g= g.prior
    let get_new_states g= g.new_states
    let get_norm_rules g= g.norm_rules
    let set_norm_rules lrules (g: gamma_content)= {prior= g.prior; new_states= g.new_states;
						   norm_rules= lrules;
						   strategy=g.strategy}
    let set_new_states state_ops (g: gamma_content)= {prior= g.prior; new_states= state_ops;
						   norm_rules= g.norm_rules;
						   strategy=g.strategy}
    let get_strategy (g: gamma_content)= g.strategy
    let set_strategy (strat: string list) (g: gamma_content)= 
      {prior= g.prior; new_states= g.new_states;
       norm_rules= g.norm_rules;
       strategy= strat}

    let rec go()= 
      (Printf.fprintf !out_chan "%s" "(y/n)? ";
       flush !out_chan;
      let rep= input_line !in_chan in
	match rep with
	    "y"|"Y"|"YES"|"yes"|"Yes" -> true
	  | "n"|"N"|"NO"|"No"|"no" -> false
	  | _ -> 
	      (Printf.fprintf !out_chan "%s" "Answer by ";
	       flush !out_chan;
	       go()))
   
			 
		      
(* the lexer for adding rules interactively *)	

    let lexer = Genlex.make_lexer [",";"(";")";"->";"."; "*";"New"; "States"]

(* the parser for adding rules interactively the result is a triple (trans, prior, new_states_ops) where
   new state_ops are the new states defined for the description of the normalising rules, and prior is a SUBSET
   of trans, such that this subset is to be added to the prior field of the gamma_content *)

    exception Multiply_defined_symbol of string


    let rec parse_user_transitions alphabet state_ops= 
      let (new_trans, new_prior, new_ops)= try (parse_user_aux alphabet state_ops (lexer (Stream.of_channel !in_chan))) with
	  Multiply_defined_symbol(m) -> 
	    begin
	      Printf.fprintf !out_chan "%s" ("\nBe careful: "^m^", try again:\n");
	      flush !out_chan;
	      parse_user_transitions alphabet state_ops
	    end
	|_ ->
	   begin
	     Printf.fprintf !out_chan "%s" "\nSyntax error in declarations, try again:\n";
	     flush !out_chan;
	     parse_user_transitions alphabet state_ops
	   end
      in 
	if List.for_all Automata_type.is_normalized (Trans_type.to_list new_prior) then (new_trans, new_prior, new_ops)
	else 
	  begin
	    Printf.fprintf !out_chan "%s" "\nUser prioritary transtions should be normalized!! \ni.e. of the form f(q1, ..., qn) -> q' or a -> q' if a is a constant tymbol\nTry again:\n";
	    flush !out_chan;
	    parse_user_transitions alphabet state_ops
	  end

	  
    and parse_user_aux alphabet state_ops=
      parser
	[< 'Genlex.Kwd "New"; 'Genlex.Kwd "States"; new_state_ops= State_set_type.parse_ops;
	   'Genlex.Kwd ".";
	   (trans, prior)= parse_transitions alphabet (Alphabet_type.union state_ops new_state_ops) 
	                     Trans_type.empty Trans_type.empty>] -> 
	  if Alphabet_type.disjoint alphabet new_state_ops then (trans, prior, new_state_ops)
	  else raise (Multiply_defined_symbol "states and symbols are not disjoint")
      | [< (trans, prior) = parse_transitions alphabet state_ops Trans_type.empty Trans_type.empty>] 
	                        -> (trans, prior, Alphabet_type.new_alphabet)
	  
    and (parse_transitions: alphabet -> alphabet -> Trans_type.t -> Trans_type.t -> Genlex.token Stream.t
	   -> (Trans_type.t * Trans_type.t))= 
	   function alphabet -> function state_ops -> function (trans: Trans_type.t) -> function (prior: Trans_type.t) -> parser
	[< 'Genlex.Kwd "." >] -> (trans, prior)
      |	[< 'Genlex.Kwd "*"; one_prior= Trans_type.parse_ground_special_rule alphabet state_ops;
	   remainder= parse_transitions alphabet state_ops (Trans_type.add one_prior trans) 
			(Trans_type.add one_prior prior)>] -> remainder
      | [< one_trans= Trans_type.parse_ground_special_rule alphabet state_ops;
	   remainder=  parse_transitions alphabet state_ops (Trans_type.add one_trans trans) 
			prior>] -> remainder



(* Special normalisation for gamma: rules with variable lhs like: x -> a 
   are applied only once to prevent non-termination *)

   let gamma_norm (term: Term.t) (trs: TRS_type.t)= 
     let rec aux trs=
       if TRS_type.is_empty trs 
       then (false, [])
       else 
	 let first_rule= TRS_type.first trs in
	 let lhs= TRS_type.lhs first_rule in
	 let rhs= TRS_type.rhs first_rule in
	   if Term.is_variable lhs then (true, [rhs])
	   else aux (TRS_type.remainder trs)
     in
     let (some_var_lhs, value)= aux trs in
       if some_var_lhs then List.hd value
       else TRS_type.bottom_up_norm trs term

      
(* look for the normalisation of a term w.r.t. to partially instanciated rewrite rules
   the result of [norm_term t trs] is a couple (t', l) where t' is a
   normalised term with the corresponding list of transitions used to normalise t in to t' *)

    let rec norm_term trs t=
      match t with
	  Special(t2) -> (t, [])
	| Const(_) -> 
	    let norm= gamma_norm t trs in
	      if Term.equal t norm then (t, [])
	      else (norm, [Trans_type.new_rule t norm])
	| Fsym(f, largs) -> 
	    let norm_args_and_trans= List.map (norm_term trs) largs in
	    let (normed_args, normed_trans)= List.split norm_args_and_trans in
	    let flat_trans= List.flatten normed_trans in
	    let not_top_normed= Fsym(f, normed_args) in
	    let top_normed= gamma_norm not_top_normed trs in
	      if Term.equal not_top_normed top_normed 
	      then (not_top_normed, flat_trans) 
	      else (top_normed, (Trans_type.new_rule not_top_normed top_normed)::flat_trans)
	| _ -> raise (Badly_formed_term ("Term "^(Term.to_string t)^" is not a valid configuration"))


(* Application of a substitution to a trs *)

    let rec trs_apply (s: subst) (r: transition_table)= 
      if TRS_type.is_empty r then r
      else
	let first= TRS_type.first r in 
	  TRS_type.add (TRS_type.new_rule 
			  (Term.apply s (TRS_type.lhs first))
			  (Term.apply s (TRS_type.rhs first)))
	    (trs_apply s (TRS_type.remainder r))
			  
			  
      

(* normalistion of a new transition [trans] w.r.t. a [norm_rule] *)

    let norm_with_one_rule trans norm_rule=
      try (let sol1= Term.matching (TRS_type.lhs norm_rule.pat)(TRS_type.lhs trans) in
	   let sol2= Term.matching (TRS_type.rhs norm_rule.pat)(TRS_type.rhs trans) in
	   let combined_subst= Term.coherent sol1@sol2 in
	   let substituted_rules= trs_apply combined_subst norm_rule.modifs in
	   let (new_config, new_trans)= norm_term substituted_rules (TRS_type.lhs trans) in
	     if Automata_type.is_state_config new_config then
	       new_trans
	     else 
	     (Trans_type.new_rule new_config (Trans_type.rhs trans))::new_trans)
      with Term.Terms_do_not_match(_,_) -> [trans]


(* normalisation of a new transition [trans] w.r.t. a list of [norm_rule] *)

    let rec norm_with_norm_rules trans (nrl: norm_rule list)=
      match nrl with
	  [] -> (false, Trans_type.list_to_trs [trans])
	| one_trans::lrem -> 
	    let res= norm_with_one_rule trans one_trans in
	      match res with
		  [t] -> 
		    if Trans_type.rule_equal t trans
		    then norm_with_norm_rules trans lrem
		    else (true, Trans_type.list_to_trs res)
		| _ -> (true, Trans_type.list_to_trs res)



(* the gamma function used for approximation of new transitions obtained during completion *)
(* After an application of one step of gamma, the gamma_content is updated, state_ops contains all initial state
   operators as well as new state operators given by the user or constructed automatically as new states. 'i' has 
   been incremented if automatic new states has been given *)

    let gamma (ltrans: transition_table) 
      (delta: (symbol, (state, rule list) folder) folder)  gamma_content alphabet state_ops i=
      
      let initial_strategy= get_strategy gamma_content in

   (* the general function applying normalising strategy *)

(*      let rec norm_strat strategy (trans_to_normalise: transition_table) 
	gamma_content state_ops i (normalised_trans: transition_table)= *)
      let rec (norm_strat: string list -> 
		 transition_table-> gamma_content-> alphabet-> int-> transition_table ->
		   (transition_table * gamma_content * alphabet * int))=
		 function strategy -> function trans_to_normalise -> function gamma_content ->
		   function state_ops -> function i -> function normalised_trans->
	let first= (Trans_type.first trans_to_normalise) in
	let lhs= Trans_type.lhs first in
	let lrem= (Trans_type.remainder trans_to_normalise) in
	  match strategy with 
	      [] ->  (* if strategy has been has been entirely executed, 
			we restart from the initial strategy *)
		
		norm_strat initial_strategy trans_to_normalise 
		  gamma_content state_ops i normalised_trans
		
	    (* Normalisation with prioritary rules: prior strategy *)
		  
	    | "prior"::rem_strat -> 
		let prior_normed= TRS_type.bottom_up_norm gamma_content.prior lhs in
		if not (Term.equal lhs prior_normed) 
		then 
		  begin
		    let the_rhs= (Trans_type.rhs first) in
		    let prior_normed_trans= Trans_type.new_rule prior_normed the_rhs in
		      Printf.fprintf !out_chan "%s" ("Prior normalisation simplifies the transition into:\n\n"^(Trans_type.rule_to_string prior_normed_trans)^"\n\n\n");
		      flush !out_chan;

(* if prior normalisation modifies the current non-normalised transition, we re-start from
   the begining with [initial_strategy] *)

		      trivial initial_strategy
			(Trans_type.add prior_normed_trans lrem)
			gamma_content
			state_ops i normalised_trans
		  end
		else
		  (* if this strategy operator has no effect we use continue with the next one *)
		  norm_strat rem_strat trans_to_normalise gamma_content state_ops i 
		    normalised_trans
		
	  | "norm_rules"::rem_strat ->
	      let (modif, new_trans)= norm_with_norm_rules first (get_norm_rules gamma_content) in
		if modif 
		then
		  begin
		    Printf.fprintf !out_chan "%s" ("Norm rules simplified the transition into:\n\n"^(Trans_type.to_string new_trans)^"\n\n\n");
		    flush !out_chan;
		  (* if [norm_rules] normalisation modifies the current non-normalised 
		     transition, we re-start from the begining with [initial_strategy] *)
		  
		    trivial initial_strategy (Trans_type.union new_trans lrem)
		      gamma_content 
		      (* we add the new states used in the [norm_rules] *)
		      (Alphabet_type.union (get_new_states gamma_content) state_ops) i
		      normalised_trans
		  end
		else
		  norm_strat rem_strat trans_to_normalise gamma_content state_ops i 
		    normalised_trans

	  | "manual_conf"::rem_strat -> 
	      begin
		Printf.fprintf !out_chan "%s" 
		  "Do you want to give by hand some rules to normalize the transition? ";
		flush !out_chan;
		if go() then 
		  norm_strat ("manual"::rem_strat) trans_to_normalise gamma_content state_ops i 
		    normalised_trans
		else norm_strat rem_strat trans_to_normalise gamma_content state_ops i 
		    normalised_trans
	      end

	  | "manual"::rem_strat -> 
	      (* Normalisation with interactively given user rules *)
	      begin
		Printf.fprintf !out_chan "%s" "Use key word 'New States' followed by the names of the new states ended by a dot '.'(optional)\n then give a sequence of transitions ended by a dot '.'\nAdd a star '*' before transitions you want to add to the prioritary set.\nThe prioritary transitions should be normalized!!\n"; 
		flush !out_chan;
		let (new_trans, new_prior, new_ops)= parse_user_transitions alphabet state_ops in
		let user_normed= TRS_type.bottom_up_norm new_trans lhs in
		let user_normed_trans= Trans_type.new_rule user_normed (Trans_type.rhs first) in
		let updated_state_ops= Alphabet_type.union new_ops state_ops in
		let new_trans_set= Trans_type.add user_normed_trans 
				     (Trans_type.union new_trans lrem) in
		let new_gamma_content= add_prior_to_gamma new_prior gamma_content in
		  if Term.equal lhs user_normed 
		  then 
		    norm_strat rem_strat new_trans_set new_gamma_content updated_state_ops i 
		      normalised_trans
		  else 
		    begin
		      Printf.fprintf !out_chan "%s" ("Normalisation simplifies the transition into: "^(Trans_type.rule_to_string user_normed_trans)^"\n\n\n");
		      flush !out_chan;
		      trivial initial_strategy new_trans_set new_gamma_content 
			updated_state_ops i normalised_trans
		    end
	      end
		   
	  | "auto_conf"::rem_strat -> 
	      begin
		Printf.fprintf !out_chan "%s" 
		  "Do you want to use automatic normalisation by NEW states to normalise the transition? ";
		flush !out_chan;
		if go() then 
		  norm_strat ("auto"::rem_strat) trans_to_normalise gamma_content state_ops i 
		    normalised_trans
		else norm_strat rem_strat trans_to_normalise gamma_content state_ops i 
		    normalised_trans
	      end
	      
	  | "auto"::rem_strat -> 
	      (* automatic normalisation by new states *)
	      let (norm_with_new, new_i, auto_new_ops) = 
		Automata_type.normalize (Trans_type.list_to_trs [first])  
		  (Trans_type.list_to_trs (Automata_type.bi_folder_flatten delta)) "qnew" i in
		trivial initial_strategy (Trans_type.union norm_with_new lrem) gamma_content
		  (Alphabet_type.union_fast auto_new_ops state_ops) new_i normalised_trans
	  | x::rem_strat -> raise (Strategy_error ("Unknown normalisation operator: "^x)) 
	      
			


(* the function retrieving transitions already normalised and epsilon transitions to 
   be automatically normalised. If the transition is not trivially normalisable, we use the 
   previous function i.e., [norm_strat] *)

       and trivial strategy (trans_to_normalise: transition_table) gamma_content state_ops i 
	(normalised_trans: transition_table)=
		   
	if Trans_type.is_empty trans_to_normalise then (normalised_trans, gamma_content, state_ops, i)
	else 
	  let first= (Trans_type.first trans_to_normalise) in
	  let lrem= (Trans_type.remainder trans_to_normalise) in
	    begin
	      Printf.fprintf !out_chan "%s" ("Adding transition:\n\n "^
					     (Trans_type.rule_to_string first)^"\n\n");
	      flush !out_chan;
	      if Automata_type.is_normalized first 
	      then  

		(* ------------- case of a transition already normalised --------------- *)

		begin
		  Printf.fprintf !out_chan "%s" " ... already normalised!\n\n\n";
		  flush !out_chan;
		  trivial strategy lrem gamma_content state_ops i
		    (Trans_type.add first normalised_trans)
		end
	      else 
		let lhs= Trans_type.lhs first in
       		  if Automata_type.is_state_config lhs then

		    (* ------ case of an epsilon transition we directly normalize it ---- *)
		    
		    begin
		      Printf.fprintf !out_chan "%s" " ... automatically normalized.\n\n\n";
		      flush !out_chan;
		      let copy= Automata_type.normalize_epsilon lhs 
				  (Trans_type.rhs first) 
				  (Trans_type.list_to_trs 
				     (Automata_type.bi_folder_flatten delta)) in
			trivial strategy lrem gamma_content state_ops i 
			  (Trans_type.union copy normalised_trans)
		    end
		  else
		    (* case where the transition is not necessary i.e. it is already covered
		       by the rules of the tree automaton *)
		    if Automata_type.is_recognized_into lhs (Trans_type.rhs first) delta 
		    then
		      begin
			Printf.fprintf !out_chan "%s" " ... covered by current automaton.\n\n\n";
			flush !out_chan;
			trivial strategy lrem gamma_content state_ops i normalised_trans
		      end
		    else
		      
		    (* -------the general case where the first transition has to be normalised 
		       i.e. it is not trivial: we use the normalisation strategy --------- *)
		      norm_strat strategy trans_to_normalise gamma_content 
			state_ops i normalised_trans
	    end

	    in
	      match initial_strategy with
		  [] -> raise (Strategy_error ("Cannot start gamma function an empty strategy!"))
		| _ -> trivial initial_strategy ltrans gamma_content state_ops i Trans_type.empty


(* We check that rules are well formed w.r.t. [alphabet] and [state_ops], and moreover that 
   in the transition table of the rhs of the norm rule [nrule], that every rhs is either a variable
   occurring in the rule of the lhs of [nrule] or a state *)

    let check_norm_rule (nrule: norm_rule) alphabet state_ops = 
      let lhs_vars= List.append (Term.list_variables (TRS_type.lhs nrule.pat)) 
		      (Term.list_variables (TRS_type.rhs nrule.pat)) in
      let list_modifs= TRS_type.to_list nrule.modifs in
      let _= List.for_all (function x -> 
			    let rhs= TRS_type.rhs x in
			    let lhs= TRS_type.lhs x in
			    let _= Term.check_special alphabet state_ops lhs in
			    let _= Term.check_special alphabet state_ops rhs in
			      match rhs with
				  Special(_) -> true
				| Var(v) -> 
				    if List.mem v lhs_vars 
				    then true
				    else raise (Error_in_parsing ("Variable "^(Term.to_string rhs)^" do not occur in "^(TRS_type.rule_to_string nrule.pat)))
				| _ ->  raise (Error_in_parsing ("Right-hand side of rule "^(TRS_type.rule_to_string x)^" is neither a state nor a variable!"))) list_modifs in nrule
				  

(* parsing of gamma content in a specification *)
 
    let rec parse alphabet variable_set= parser
	[< 'Genlex.Kwd "States"; 
	   state_ops = State_set_type.parse_ops;
	   norm_rules= parse_rules alphabet state_ops variable_set>]  ->
	  if Alphabet_type.disjoint alphabet state_ops 
	  then set_new_states state_ops (set_norm_rules norm_rules empty)
	  else raise (Multiply_defined_symbol "states and symbols are not disjoint in the specification")
	  
    and parse_rules alphabet state_ops variable_set =
      parser
	  [< 'Genlex.Kwd "Rules";
	     rules= parse_rules_aux alphabet state_ops variable_set>] -> rules
	| [< >] -> raise (Error_in_parsing "Missing 'Rules' field in the approximation")

    and parse_rules_aux alphabet state_ops variable_set =
      parser
	  [< rule= parse_one_rule alphabet state_ops variable_set; 
	     remainder= parse_rules_aux alphabet state_ops variable_set>] ->
	    (check_norm_rule rule alphabet state_ops)::remainder
	| [< >] -> []

    and parse_one_rule alphabet state_ops variable_set= parser
	[< 'Genlex.Kwd "[";
	   patrn = TRS_type.parse_special_rule alphabet state_ops variable_set;
	   'Genlex.Kwd "]"; 'Genlex.Kwd "->"; 'Genlex.Kwd "[";
	   rules= TRS_type.parse_special alphabet state_ops variable_set;
	   'Genlex.Kwd "]" >] -> {pat=patrn; modifs= rules}


(* pretty print *)

    let rule_to_string norm_rule=
      ("[ "^(TRS_type.rule_to_string norm_rule.pat)^"] ->\n[ "^
       (TRS_type.to_string norm_rule.modifs)^" ]")
       
    let to_string (g: gamma_content)=
      ("States "^(Alphabet_type.to_string (get_new_states g))^"\n\n"^
       "Rules\n"^(List.fold_right (function x -> function y -> (rule_to_string x)^"\n\n"^y)
		    (get_norm_rules g) "")^"\n\n")


end

