(* Timbuk - A Tree Automata Library (for Ocaml)
   Copyright (C) 2000-2001 Thomas Genet and Valrie Viet Triem Tong

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Last modification 28/02/01
*)

(*i Be careful! Some comments of this file are formatted for ocamlweb. *)

(*i*)
open Common
(*i*)
   (* This is the interface for the gamma approximation module. The role
      of the gamma function is to realize an approximate normalisation of
      new transition in order to make completion terminate as often as
      possible. This function is defined thanks to a database called
      [gamma_content] which can be enriched during completion. The role of
      this database is to store rules of normalisation, events encountered
      during the completion, ... or anything necessary for making good
      heuristical choices for normalisation of new critical pairs. The gamma
      content as well as the gamma function itself can be defined by
      the user. However we propose one simple gamma function here
      which is very simple and only take into account a set of
      prioritary transitions (see [tutorial_compl.mlml] for more
      details on prioritary transitions) in the [gamma_content]
      database. This set of prioritary transitions is being enriched
      by the user during completion and used for normalising
      transitions before giving them to an interactive normalisation process.

 *)


module Gamma 
  (Symbol_type: PRINTABLE_TYPE)
  (Alphabet_type: ALPHABET_TYPE with type symbol= Symbol_type.t)
  (Variable_set_type: VARIABLE_SET_TYPE)
  (Term: TERM_TYPE with type symbol= Symbol_type.t 
		   and type alphabet= Alphabet_type.t
		   and type variable_set= Variable_set_type.t)
  (TRS_type: TRS_TYPE with type alphabet= Alphabet_type.t 
		      and type term= Term.t
		      and type variable_set= Variable_set_type.t)
  (State_set_type: STATE_SET_TYPE with type state= Term.t
				  and type alphabet= Alphabet_type.t 
				  and type symbol= Symbol_type.t)
  (Automata_type: AUTOMATA_TYPE with type symbol= Symbol_type.t 
				and type alphabet= Alphabet_type.t 
				and type term= Term.term 
				and type transition_table= TRS_type.t 
				and type rule= TRS_type.rule):
sig
  type alphabet= Alphabet_type.t
  type automaton= Automata_type.t
  type state = Automata_type.state
  type transition_table= TRS_type.t
  type transition= TRS_type.rule
  type gamma_content
  type variable_set= Variable_set_type.t
  type ('a, 'b) folder= ('a, 'b) Automata_type.folder
  type symbol= Symbol_type.t
  type rule= TRS_type.rule

  (*s the empty gamma content of gamma approximation *)
  val empty : gamma_content

  (*s adding a prioritary transition table to the gamma content *) 
  val add_prior_to_gamma :
    transition_table -> gamma_content -> gamma_content

  (*s setting the gamma content to a new prior transition table *)
  val set_prior_in_gamma :
    transition_table -> gamma_content -> gamma_content

  (*s getting back the transition table from the gamma content *)
  val get_prior : gamma_content -> transition_table

  (*s getting the current strategy of gamma *)
  val get_strategy : gamma_content-> string list
      
  (*s setting the strategy in the gamma content *)    
  val set_strategy : string list -> gamma_content -> gamma_content
 
  (*s the parsing function for gamma content *)
  val parse :
    alphabet ->
      variable_set -> Genlex.token Stream.t -> gamma_content

  (*s the pretty printing function for gamma content *)
  val to_string :
    gamma_content -> string

  (*s The gamma function used for approximation of new transitions
  obtained during completion.  The [new_trans] are the new transitions
  to be normalised, [delta] is the transition table of the current
  completion step (stored into a folder for a better efficiency), 
  [a] is the alphabet and [stops] are the state 
  operators set on which transitions are defined, and finally [i] is
  the index of the lastly created new state. After an application of
  one step of gamma, the gamma content is updated, the new state
  operators contains all initial state operators as well as new state
  operators given by the user or constructed automatically as new
  states. The index has been incremented if automatic new states have been
  constructed *)


  val gamma :
    new_trans:transition_table ->
      delta:(symbol, (state, rule list) folder) folder ->
        gc:gamma_content ->
          a:alphabet ->
            stop:alphabet ->
              ind:int -> transition_table * gamma_content * alphabet * int

(*i*)
  val go: unit -> bool
(*i*)

  (*s the in and out channel used for user interaction *)
  val out_chan : out_channel ref
  val in_chan : in_channel ref


end
